import 'package:binarize/binarize.dart';

class _Map<K, V> extends PayloadType<Map<K, V>> {
  const _Map(this.keyType, this.valueType, this.lengthType);

  final PayloadType<K> keyType;

  final PayloadType<V> valueType;

  final PayloadType<int> lengthType;

  @override
  int length(Map<K, V> value) {
    var currentOffset = lengthType.length(value.length);
    for (final k in value.keys) {
      currentOffset += keyType.length(k);
      currentOffset += valueType.length(value[k] as V);
    }
    return currentOffset;
  }

  @override
  Map<K, V> get(ByteData data, int offset) {
    final length = lengthType.get(data, offset);
    var currentOffset = offset + lengthType.length(length);

    final result = <K, V>{};
    for (var i = 0; i < length; i++) {
      final k = keyType.get(data, currentOffset);
      currentOffset += keyType.length(k);

      final v = valueType.get(data, currentOffset);
      currentOffset += valueType.length(v);

      result[k] = v;
    }
    return result;
  }

  @override
  void set(Map<K, V> value, ByteData data, int offset) {
    lengthType.set(value.length, data, offset);
    var currentOffset = offset + lengthType.length(value.length);

    for (final k in value.keys) {
      keyType.set(k, data, currentOffset);
      currentOffset += keyType.length(k);

      final v = value[k] as V;
      valueType.set(v, data, currentOffset);
      currentOffset += valueType.length(v);
    }
  }
}

/// Store a map of [K] and [V] as binary data.
///
/// It stores the length of the map first by using the [lengthType] and then
/// stores each key value pair in the map by looping through the keys, storing
/// the key first using [keyType] and then the value of the key using
/// [valueType].
///
/// It has a byte length of:
/// ```dart
/// lengthType.length(map.length) + map.entries.fold(0,
///   (p, e) => p + keyType.length(e.key) + valueType.length(e.value),
/// )
/// ```
///
/// Usage:
///
/// ```dart
/// payload.set(map(string8, uint16), {'1': 1, '2': 2, '3': 3, '4': 4});
/// ```
PayloadType<Map<K, V>> map<K, V>(
  PayloadType<K> keyType,
  PayloadType<V> valueType, {
  PayloadType<int> lengthType = uint8,
}) {
  return _Map(keyType, valueType, lengthType);
}
