// ignore_for_file: prefer_const_constructors
// ignore_for_file: avoid_equals_and_hash_code_on_mutable_classes

import 'package:binarize/binarize.dart';
import 'package:test/test.dart';

import '../helpers/helpers.dart';

class Person {
  const Person({
    required this.firstName,
    required this.lastName,
    required this.age,
    this.children = const [],
  });

  final String firstName;

  final String lastName;

  final int age;

  final List<Person> children;

  @override
  bool operator ==(Object other) {
    return other is Person &&
        other.firstName == firstName &&
        other.lastName == lastName &&
        other.age == age &&
        other.children.every(children.contains);
  }

  @override
  int get hashCode => Object.hashAll([firstName, lastName, age, children]);

  @override
  String toString() =>
      '''Person(firstName: $firstName, lastName: $lastName, age: $age, children: $children)''';
}

class PersonContract extends BinaryContract<Person> implements Person {
  const PersonContract()
      : super(const Person(firstName: '', lastName: '', age: 42));

  @override
  Person order(Person contract) {
    return Person(
      firstName: contract.firstName,
      lastName: contract.lastName,
      age: contract.age,
      children: contract.children,
    );
  }

  @override
  String get firstName => type(string8, (o) => o.firstName);

  @override
  String get lastName => type(string8, (o) => o.lastName);

  @override
  int get age => type(uint8, (o) => o.age);

  @override
  List<Person> get children =>
      type(list(personContract, lengthType: uint8), (o) => o.children);
}

final personContract = PersonContract();

void main() {
  group('$BinaryContract', () {
    final bytes = [4, 74, 111, 104, 110, 3, 68, 111, 101, 42, 0];
    final person = Person(firstName: 'John', lastName: 'Doe', age: 42);

    test(
      'turns instance to binary using the contract',
      expectGet(personContract, bytes, person),
    );

    test(
      'turns binary to an instance using the contract',
      expectSet(personContract, person, bytes),
    );

    test(
      'retrieve length based on the contract',
      expectLength(personContract, person, bytes.length),
    );

    test('supports self-referencing types', () {
      final mom = Person(
        firstName: 'Jane',
        lastName: 'Doe',
        age: 89,
        children: [person],
      );
      final bytes = [
        4, 74, 97, 110, 101, 3, 68, 111, //
        101, 89, 1, 4, 74, 111, 104,
        110, 3, 68, 111, 101, 42, 0
      ];

      expectGet(personContract, bytes, mom)();
      expectSet(personContract, mom, bytes)();
      expectLength(personContract, mom, bytes.length)();
    });
  });

  group('boolean', () {
    test('returns true', expectGet(boolean, [1], true));
    test('returns false', expectGet(boolean, [0], false));

    test('sets bytes to true', expectSet(boolean, true, [1]));
    test('sets bytes to false', expectSet(boolean, false, [0]));

    test('byte length is 1 for true', expectLength(boolean, true, 1));
    test('byte length is 1 for false', expectLength(boolean, false, 1));
  });

  group('$Bytes', () {
    test('returns list of bytes', expectGet(Bytes(3), [1, 2, 3], [1, 2, 3]));

    test(
      'sets bytes to bytes',
      expectSet(Bytes(3), [1, 2, 3], [1, 2, 3]),
    );

    test(
      'byte length is equal to list length',
      expectLength(Bytes(3), [1, 2, 3], 3),
    );
  });

  group('flags', () {
    final bytes = [5];
    final testFlags = [true, false, true, false, false, false, false, false];

    test('returns correct flags from byte', expectGet(flags, bytes, testFlags));

    test('sets byte to represent flags', expectSet(flags, testFlags, bytes));

    test('byte length is 1', expectLength(flags, testFlags, 1));

    test('throws assertion when a list of more than 8 booleans is passed', () {
      expect(
        () => flags.set([...testFlags, true], ByteData(0), 0),
        throwsAssertionMessage('Flags can only be 8 bits long'),
      );
    });
  });

  group('float32', () {
    final bytes = [63, 192, 0, 0];

    test('returns 1.5', expectGet(float32, bytes, 1.5));

    test('sets bytes to 1.5', expectSet(float32, 1.5, bytes));

    test('byte length is 4', expectLength(float32, 1.5, 4));
  });

  group('float64', () {
    final bytes = [63, 248, 0, 0, 0, 0, 0, 0];

    test('returns 1.5', expectGet(float64, bytes, 1.5));

    test('sets bytes to 1.5', expectSet(float64, 1.5, bytes));

    test('length is 8', expectLength(float64, 1.5, 8));
  });

  group('int8', () {
    test('returns -8', expectGet(int8, [248], -8));
    test('returns 8', expectGet(int8, [8], 8));

    test('sets bytes to -8', expectSet(int8, -8, [248]));
    test('sets bytes to 8', expectSet(int8, 8, [8]));

    test('byte length is 1', expectLength(int8, 8, 1));
  });

  group('int16', () {
    test('returns -16', expectGet(int16, [255, 240], -16));
    test('returns 16', expectGet(int16, [0, 16], 16));

    test('sets bytes to -16', expectSet(int16, -16, [255, 240]));
    test('sets bytes to 16', expectSet(int16, 16, [0, 16]));

    test('byte length is 2', expectLength(int16, 16, 2));
  });

  group('int32', () {
    test('returns -32', expectGet(int32, [255, 255, 255, 224], -32));
    test('returns 32', expectGet(int32, [0, 0, 0, 32], 32));

    test('sets bytes to -32', expectSet(int32, -32, [255, 255, 255, 224]));
    test('sets bytes to 32', expectSet(int32, 32, [0, 0, 0, 32]));

    test('byte length is 4', expectLength(int32, 32, 4));
  });

  group('int64', () {
    test(
      'returns -64',
      expectGet(int64, [255, 255, 255, 255, 255, 255, 255, 192], -64),
    );
    test('returns 64', expectGet(int64, [0, 0, 0, 0, 0, 0, 0, 64], 64));

    test(
      'sets bytes to -64',
      expectSet(int64, -64, [255, 255, 255, 255, 255, 255, 255, 192]),
    );
    test('sets bytes to 64', expectSet(int64, 64, [0, 0, 0, 0, 0, 0, 0, 64]));

    test('byte length is 8', expectLength(int64, 64, 8));
  });

  group('list', () {
    final bytes = [0, 0, 0, 3, 0, 1, 0, 2, 0, 3];

    test(
      'returns a list',
      expectGet(list(uint16), bytes, [1, 2, 3]),
    );

    test('sets bytes to a list', expectSet(list(uint16), [1, 2, 3], bytes));

    test(
      'byte length is equal to length type + length of all element in the list',
      expectLength(list(uint16, lengthType: uint8), [1, 2, 3], 1 + 3 * 2),
    );
  });

  group('map', () {
    final data = {'1': 1, '2': 2, '3': 3, '4': 4};
    final bytes = [4, 1, 49, 0, 1, 1, 50, 0, 2, 1, 51, 0, 3, 1, 52, 0, 4];

    test('returns a map', expectGet(map(string8, uint16), bytes, data));

    test('sets bytes to a map', expectSet(map(string8, uint16), data, bytes));

    test(
      'byte length is equal to length type + length of all pairs in the map',
      expectLength(
        map(string8, uint16),
        data,
        1 + data.length + 2 * data.length + data.length,
      ),
    );
  });

  group('nullable', () {
    test('returns null', expectGet(nullable(uint8), [0], null));

    test('stores null', expectSet(nullable(uint8), null, [0]));

    test(
      'byte length is 1 if value is null',
      expectLength(nullable(uint8), null, 1),
    );

    test(
      'byte length is 1 + value byte length if value is not null',
      expectLength(nullable(uint8), 1, 2),
    );

    test(
      'returns value if value is not null',
      expectGet(nullable(uint8), [1, 2], 2),
    );

    test(
      'sets value if value is not null',
      expectSet(nullable(uint8), 2, [1, 2]),
    );
  });

  group('$RawList', () {
    const data = ['1', '2', '3'];
    final bytes = [1, 49, 1, 50, 1, 51];

    test(
      'returns raw list of string',
      expectGet(RawList(string8, amount: 3), bytes, data),
    );

    test(
      'sets bytes to raw list of string',
      expectSet(RawList(string8, amount: 3), data, bytes),
    );

    test(
      'bytes length is equal to list length + type length',
      expectLength(RawList(string8, amount: 3), ['1', '2', '3'], 6),
    );
  });

  group('string8', () {
    const data = 'Hello World';
    final bytes = [
      11, 72, 101, 108, 108, //
      111, 32, 87, 111,
      114, 108, 100
    ];

    test('returns "$data"', expectGet(string8, bytes, data));

    test('sets bytes to "$data"', expectSet(string8, data, bytes));

    test(
      'byte length is dynamic based on string length',
      expectLength(string8, data, 1 + data.length),
    );

    test(
      'has a minimal length of 1, indicating string length',
      expectLength(string8, '', 1),
    );
  });

  group('string16', () {
    const data = 'Hello World';
    final bytes = [
      0, 11, 72, //
      101, 108, 108,
      111, 32, 87, 111,
      114, 108, 100
    ];

    test('returns "$data"', expectGet(string16, bytes, data));

    test('sets bytes to "$data"', expectSet(string16, data, bytes));

    test(
      'byte length is dynamic based on string length',
      expectLength(string16, data, 2 + data.length),
    );

    test(
      'has a minimal length of 2, indicating string length',
      expectLength(string16, '', 2),
    );
  });

  group('string32', () {
    const data = 'Hello World';
    final bytes = [
      0, 0, 0, 11, //
      72, 101, 108, 108,
      111, 32, 87, 111,
      114, 108, 100
    ];

    test('returns "$data"', expectGet(string32, bytes, data));

    test('sets bytes to "$data"', expectSet(string32, data, bytes));

    test(
      'byte length is dynamic based on string length',
      expectLength(string32, data, 4 + data.length),
    );

    test(
      'has a minimal length of 4, indicating string length',
      expectLength(string32, '', 4),
    );
  });

  group('string64', () {
    const data = 'Hello World';
    final bytes = [
      0, 0, 0, 0, //
      0, 0, 0, 11,
      72, 101, 108, 108,
      111, 32, 87, 111,
      114, 108, 100
    ];

    test('returns "$data"', expectGet(string64, bytes, data));

    test('sets bytes to "$data"', expectSet(string64, data, bytes));

    test(
      'byte length is dynamic based on string length',
      expectLength(string64, data, 8 + data.length),
    );

    test(
      'has a minimal length of 8, indicating string length',
      expectLength(string64, '', 8),
    );
  });

  group('uint8', () {
    test('returns 8', expectGet(uint8, [8], 8));

    test('sets bytes to 8', expectSet(uint8, 8, [8]));

    test('byte length is 1', expectLength(uint8, 8, 1));
  });

  group('uint16', () {
    test('returns 16', expectGet(uint16, [0, 16], 16));

    test('sets bytes to 16', expectSet(uint16, 16, [0, 16]));

    test('byte length is 2', expectLength(uint16, 16, 2));
  });

  group('uint32', () {
    test('returns 32', expectGet(uint32, [0, 0, 0, 32], 32));

    test('sets bytes to 32', expectSet(uint32, 32, [0, 0, 0, 32]));

    test('byte length is 4', expectLength(uint32, 32, 4));
  });

  group('uint64', () {
    test('returns 64', expectGet(uint64, [0, 0, 0, 0, 0, 0, 0, 64], 64));

    test('sets bytes to 64', expectSet(uint64, 64, [0, 0, 0, 0, 0, 0, 0, 64]));

    test('byte length is 8', expectLength(uint64, 64, 8));
  });
}

void Function() expectGet<T>(PayloadType<T> type, List<int> input, T expected) {
  return () {
    return expect(
      type.get(ByteData.view(Uint8List.fromList(input).buffer), 0),
      equals(expected),
    );
  };
}

void Function() expectSet<T>(PayloadType<T> type, T input, List<int> expected) {
  return () {
    final setData = ByteData(expected.length);
    type.set(input, setData, 0);

    return expect(
      setData.buffer.asUint8List(),
      equals(Uint8List.fromList(expected)),
    );
  };
}

void Function() expectLength<T>(PayloadType<T> type, T input, int expected) {
  return () => expect(type.length(input), equals(expected));
}
